
#include <stdio.h>
#include <string.h>
#include <sys/timeb.h>

#include "template.h"

class Instance {
protected:
  struct Vec { int vaha, cena; };
  int ID,n,M;
  CArray<Vec> veci;
public:
  Instance() { }
  Instance(const char *line) {
    int read;
    if (sscanf(line,"%d %d %d%n",&ID,&n,&M,&read)<3) throw "Chyba pri cteni souboru";
    line += read;
    veci.Init(n);
    for (int i=0; i<n; i++) {
      if (sscanf(line,"%d %d%n",&veci[i].vaha,&veci[i].cena,&read)<2) throw "Chyba pri cteni souboru";
      line += read;
    }
  }
  int Pocet() {
    return n;
  }
  int Limit() {
    return M;
  }
  int Id() {
    return ID;
  }
  int Vaha(int i) {
    return veci[i].vaha;
  }
  int Cena(int i) {
    return veci[i].cena;
  }
  int CenaVsechVeci() {
    int celkem = 0;
    for (int i=0; i<n; i++) celkem += Cena(i);
    return celkem;
  }
};

class Konfigurace {
protected:
  Instance instance;
  CArray<bool> veci;
  int cena, vaha, vybrano;
  bool initialized;
public:
  void Init() {
    for (int i=0; i<instance.Pocet(); i++) veci[i] = false;
    cena = vaha = vybrano = 0;
    initialized = true;
  }
  bool Initialized() {
    return initialized;
  }
  Konfigurace() {
    initialized = false;
  }
  Konfigurace(Instance instance) {
    this->instance = instance;
    veci.Init(instance.Pocet());
    Init();
  }
  bool Vybrano(int vec) {
    return veci[vec];
  }
  void Pridej(int vec) {
    if (Vybrano(vec)) throw "Opakovane pridani";
    veci[vec] = true;
    cena += instance.Cena(vec);
    vaha += instance.Vaha(vec);
    vybrano++;
  }
  void Uber(int vec) {
    if (!Vybrano(vec)) throw "Opakovane ubrani";
    veci[vec] = false;
    cena -= instance.Cena(vec);
    vaha -= instance.Vaha(vec);
    vybrano--;
  }
  int Cena() {
    return cena;
  }
  int Vaha() {
    return vaha;
  }
  int PocetVybranych() {
    return vybrano;
  }
  bool Pretizen() {
    return (Vaha()>instance.Limit());
  }
  void Vypis() {
    for (int i=0; i<instance.Pocet(); i++) {
      printf(Vybrano(i) ? "1" : "0");
    }
    printf("\n");
  }
};

class Algoritmus {
protected:
  Instance instance;
  Konfigurace nejlepsi_konfigurace;
  timeb start, stop;
  unsigned long kroku;
protected:
  Algoritmus(Instance instance) : nejlepsi_konfigurace(instance) {
    this->instance = instance;
  }
  void StartTime() {
    ftime(&start);
    kroku = 0;
  }
  void StopTime() {
    ftime(&stop);
  }
  void DalsiKrok() {
    kroku++;
  }
public:
  virtual ~Algoritmus() { }
  virtual void Start() = 0;
  void VypisReseni(FILE *reseni) {
    fprintf(reseni,"%d %d %d ",instance.Id(),instance.Pocet(),nejlepsi_konfigurace.Cena());
    for (int i=0; i<instance.Pocet(); i++) {
      fprintf(reseni," %c",nejlepsi_konfigurace.Vybrano(i) ? '1' : '0');
    }
    fprintf(reseni,"\n");
    fflush(reseni);
  }
  void VypisReport(FILE *report) {
    fprintf(report,"%d %d %d\n",instance.Id(),nejlepsi_konfigurace.Cena(),kroku);
    fflush(report);
  }
  void VypisCas(FILE *cas) {
    int ms = 1000*(stop.time-start.time) + stop.millitm - start.millitm;
    fprintf(cas,"%d %d %d %d %d\n",instance.Id(),instance.Pocet(),nejlepsi_konfigurace.Cena(),ms,kroku);
    fflush(cas);
  }
};

class HrubaSila : public Algoritmus { // 1
protected:
  Konfigurace konfigurace;
public:
  HrubaSila(Instance instance) : Algoritmus(instance), konfigurace(instance) { }
  void TestNejlepsi() {
    if (konfigurace.Cena()>nejlepsi_konfigurace.Cena() && !konfigurace.Pretizen()) {
      nejlepsi_konfigurace = konfigurace;
    }
  }
  void Dalsi(int vec) {
    if (vec>=instance.Pocet() || konfigurace.Pretizen()) return;
    DalsiKrok();
    konfigurace.Pridej(vec);
    TestNejlepsi();
    Dalsi(vec+1);
    konfigurace.Uber(vec);
    TestNejlepsi();
    Dalsi(vec+1);
  }
  void Start() {
    StartTime();
    konfigurace.Init();
    nejlepsi_konfigurace.Init();
    Dalsi(0);
    StopTime();
  }
};

class HeuristikaCenaVaha : public Algoritmus { // 2
protected:
  Konfigurace konfigurace;
public:
  HeuristikaCenaVaha(Instance instance) : Algoritmus(instance), konfigurace(instance) { }
  int Najdi() {
    int nejlepsi_i = 0;
    float nejlepsi_pomer = 0;
    for (int i=0; i<instance.Pocet(); i++) {
      DalsiKrok();
      if (!konfigurace.Vybrano(i)) {
        float pomer = (float)instance.Cena(i)/instance.Vaha(i);
        if (pomer>nejlepsi_pomer) {
          nejlepsi_pomer = pomer;
          nejlepsi_i = i;
        }
      }
    }
    return nejlepsi_i;
  }
  void Start() {
    StartTime();
    konfigurace.Init();
    while (konfigurace.PocetVybranych()<instance.Pocet()) {
      DalsiKrok();
      int vec = Najdi();
      konfigurace.Pridej(vec);
      if (konfigurace.Pretizen()) {
        konfigurace.Uber(vec);
        break;
      }
    }
    nejlepsi_konfigurace = konfigurace;
    StopTime();
  }
};

class MetodaVetviHranic : public HrubaSila { // 3
public:
  MetodaVetviHranic(Instance instance) : HrubaSila(instance) { }
  void Dalsi(int vec, int potencialni_cena) {
    if (vec>=instance.Pocet() || konfigurace.Pretizen()) return;
    if (potencialni_cena<nejlepsi_konfigurace.Cena()) return;
    DalsiKrok();
    konfigurace.Pridej(vec);
    TestNejlepsi();
    Dalsi(vec+1,potencialni_cena); // S TOUTO VECI
    konfigurace.Uber(vec);
    TestNejlepsi();
    Dalsi(vec+1,potencialni_cena-instance.Cena(vec)); // BEZ TETO VECI
  }
  void Start() {
    StartTime();
    konfigurace.Init();
    nejlepsi_konfigurace.Init();
    Dalsi(0,instance.CenaVsechVeci());
    StopTime();
  }
};

class DynamickeProgramovani : public Algoritmus { // 4
protected:
  typedef CArray<Konfigurace> CacheRow;
  CArray<CacheRow> cache;
public:
  DynamickeProgramovani(Instance instance) : Algoritmus(instance) { }
  Konfigurace MaxCena(int vec, int limit) {
    if (vec<0) { Konfigurace prazdna(instance); return prazdna; }
    if (cache[vec][limit].Initialized()) return cache[vec][limit];
    DalsiKrok();
    Konfigurace lepsi = MaxCena(vec-1,limit);
    if (instance.Vaha(vec)<=limit) {
      Konfigurace pridat = MaxCena(vec-1,limit-instance.Vaha(vec));
      pridat.Pridej(vec);
      if (pridat.Cena()>=lepsi.Cena()) lepsi = pridat;
    }
    cache[vec][limit] = lepsi;
    return lepsi;
  }
  void Start() {
    StartTime();

    cache.Init(instance.Pocet());
    for (int i=0; i<cache.Count(); i++) cache[i].Init(1+instance.Limit());

    nejlepsi_konfigurace = MaxCena(instance.Pocet()-1,instance.Limit());

    StopTime();
  }
};

class HeuristikaCenaVahaTestNejdrazsiVeci : public HeuristikaCenaVaha { // 5
public:
  HeuristikaCenaVahaTestNejdrazsiVeci(Instance instance) : HeuristikaCenaVaha(instance) { }
  void Start() {
    HeuristikaCenaVaha::Start();
    int nejdrazsi_i = 0;
    int nejvyssi_cena = 0;
    for (int i=0; i<instance.Pocet(); i++) {
      DalsiKrok();
      if (instance.Cena(i)>nejvyssi_cena) {
        nejvyssi_cena = instance.Cena(i);
        nejdrazsi_i = i;
      }
    }
    Konfigurace nejdrazsi_vec(instance);
    nejdrazsi_vec.Pridej(nejdrazsi_i);
    if (nejdrazsi_vec.Pretizen()) nejdrazsi_vec.Uber(nejdrazsi_i);
    if (nejdrazsi_vec.Cena()>nejlepsi_konfigurace.Cena()) nejlepsi_konfigurace = nejdrazsi_vec;
    StopTime();
  }
};

void main(int argc, char *argv[]) {
  try {
    if (argc<=5) throw "Malo parametru";
    FILE *zadani = fopen(argv[1],"r");
    FILE *reseni = fopen(argv[2],"w");
    FILE *report = fopen(argv[3],"w");
    FILE *cas    = fopen(argv[4],"w");
    int id_algoritmu; sscanf(argv[5],"%d",&id_algoritmu);
    char line[1024];
    while (fgets(line,1024,zadani)) {
      Instance instance(line);
      Algoritmus *algoritmus = NULL;
      switch (id_algoritmu) {
        case 1: algoritmus = new HrubaSila(instance); break;
        case 2: algoritmus = new HeuristikaCenaVaha(instance); break;
        case 3: algoritmus = new MetodaVetviHranic(instance); break;
        case 4: algoritmus = new DynamickeProgramovani(instance); break;
        case 5: algoritmus = new HeuristikaCenaVahaTestNejdrazsiVeci(instance); break;
        default : throw "Spatne cislo algoritmu";
      }
      algoritmus->Start();
      algoritmus->VypisReseni(reseni);
      algoritmus->VypisReport(report);
      algoritmus->VypisCas(cas);
      delete algoritmus;
    }
    fclose(zadani);
    fclose(reseni);
    fclose(report);
    fclose(cas);
  }
  catch (const char *err) {
    printf("Exception : %s \n",err);
  }
}
